import uvicorn
import pydantic
import shap
import os
import csv
import joblib
import pandas as pd
import io
from typing import List, Optional
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from fastapi_jwt_auth import AuthJWT
from fastapi_jwt_auth.exceptions import AuthJWTException
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from fastapi import Request, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import StreamingResponse

ATS_RESOLUTION = 10

app = FastAPI(title='AIR API', version='0.1',
              description='A simple API that classifies citizens based on data')

origins = [
    "http://localhost",
    "http://localhost:8080",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

class User(pydantic.BaseModel):
    username: str
    password: str

class Settings(pydantic.BaseModel):
    authjwt_secret_key: str = "secret"
    
class JWTBearer(HTTPBearer):
    def __init__(self, auto_error: bool = True):
        super(JWTBearer, self).__init__(auto_error=auto_error)

    async def __call__(self, request: Request):
        credentials: HTTPAuthorizationCredentials = await super(JWTBearer, self).__call__(request)
        if credentials:
            if not credentials.scheme == "Bearer":
                raise HTTPException(status_code=403, detail="Invalid authentication scheme.")
            if not self.verify_jwt(request, credentials.credentials):
                raise HTTPException(status_code=403, detail="Invalid token or expired token.")
            return credentials.credentials
        else:
            raise HTTPException(status_code=403, detail="Invalid authorization code.")

    # noinspection PyBroadException
    @staticmethod
    def verify_jwt(request: Request, token: Optional[str]) -> bool:
        is_token_valid: bool = False
        auth_jwt = AuthJWT(request, token)
        try:
            payload = auth_jwt.get_raw_jwt()
        except Exception:
            payload = None
        if payload:
            is_token_valid = True
        return is_token_valid

@AuthJWT.load_config
def get_config():
    return Settings()

@app.exception_handler(AuthJWTException)
def authjwt_exception_handler(request: Request, exc: AuthJWTException):
    return JSONResponse(
        status_code=exc.status_code,
        content={"detail": exc.message}
    )

class InputData(pydantic.BaseModel):
    Gender: int
    BirthYear: int
    LoanPeriod: float
    Ats: str
    
class OutputData(pydantic.BaseModel):
    CompleteProb: float
    Compliance: int
    FallProb: float
    CompleteShapValues: dict
    FallShapValues: dict
    ClusterId: int
    ClusterValues: List
    Arguments: List

@app.get('/')
def index():
    return {'message': f'AIR API v. 0.1'}

@app.get('/user', dependencies=[Depends(JWTBearer())], tags=["login"])
def user(Authorize: AuthJWT = Depends()):
    Authorize.jwt_required()
    current_user = Authorize.get_jwt_subject()
    return {"user": current_user}

@app.get('/ats', tags=["data"])
def get_all_ats():
    return read_csv('ats.csv')

@app.get('/ex', tags=["data"])
def get_all_ex():
    return read_csv('ex.csv')

@app.get("/ats/id", tags=["data"])
def get_ats_by_id(ats_id: str):
    ats = read_csv('ats.csv')
    if ats_id in ats:
        return {"ats_id": ats_id, "ats_name": ats[ats_id]}
    raise HTTPException(status_code=404, detail="Not found")

@app.get("/ex/ex_id", tags=["data"])
def get_ex_by_id(ex_id: str):
    ex = read_csv('ex.csv')
    if ex_id in ex:
        return {"ex_id": ex_id, "ex_name": ex[ex_id]}
    raise HTTPException(status_code=404, detail="Not found")

@app.get("/cluster", tags=["data"])
def get_all_clusters():
    converter = {str(i):str for i in range(1, 20)}
    df = read_dataframe('clusters.csv', converters=converter)
    stream = io.StringIO()
    df.to_csv(stream, index = False)
    response = StreamingResponse(iter([stream.getvalue()]),
                                 media_type="text/csv")
    response.headers["Content-Disposition"] = "attachment; filename=clusters.csv"
    return response
    
@app.get("/cluster/cluster_id", tags=["data"])
def get_cluster_by_id(cluster_id: int):
    converter = {str(i):str for i in range(1, 20)}
    df = read_dataframe('clusters.csv', converters=converter)
    if df.shape[1] >= cluster_id:
        cluster = list(df.iloc[:, cluster_id])
        return {"cluster": cluster}
    raise HTTPException(status_code=404, detail="Not found")

@app.post('/login', tags=["login"])
def login(user: User, Authorize: AuthJWT = Depends()):
    if user.username != "test" or user.password != "test":
        raise HTTPException(status_code=401, detail="Bad username or password")
    access_token = Authorize.create_access_token(subject=user.username)
    refresh_token = Authorize.create_refresh_token(subject=user.username)
    return {"access_token": access_token, "refresh_token": refresh_token}

@app.post('/refresh', dependencies=[Depends(JWTBearer())], tags=["login"])
def refresh(Authorize: AuthJWT = Depends()):
    Authorize.jwt_refresh_token_required()
    current_user = Authorize.get_jwt_subject()
    new_access_token = Authorize.create_access_token(subject=current_user)
    return {"access_token": new_access_token}

@app.post('/predict', response_model=OutputData, tags=["ai"])
def predict(incoming_data: InputData):
    data = incoming_data.dict()
    
    incoming_ats = [x.strip(' ') for x in data['Ats'].split(",")]
    ats_converter = {'ats_id': str}
    ats_df = read_dataframe('ats.csv', converters=ats_converter)
    if all(x in list(ats_df['ats_id']) for x in incoming_ats) != True:
        raise HTTPException(status_code=400, detail="Ats not found, check ats list")
    
    df = prepare_data(data)
    arguments = generate_arguments(df)
    
    complete_model = read_joblib(f'complete_xgboost.joblib')
    compliance_model = read_joblib(f'compliance_xgboost.joblib')
    fall_model = read_joblib(f'fall_xgboost.joblib')
    
    df_for_complete = add_embeddings(df.copy(), 'complete')
    df_for_compliance = add_embeddings(df.copy(), 'compliance')
    df_for_fall = add_embeddings(df.copy(), 'fall')
    
    complete_prob = complete_model.predict_proba(df_for_complete).flatten()[1]
    fall_prob = fall_model.predict_proba(df_for_fall).flatten()[1]
    if complete_prob > 0.5:
        compliance_prob = compliance_model.predict_proba(df_for_compliance).flatten()[1]
    else:
        compliance_prob = 0

    cluster_id = int(df.iloc[0]['Cluster'])
    cluster_converter = {str(i):str for i in range(1, 20)}
    clusters = read_dataframe('clusters.csv', converters=cluster_converter)
    cluster_values = list(clusters.iloc[:, cluster_id])
    complete_shap_values = get_shap_values(complete_model, X_test=df_for_complete)
    fall_shap_values = get_shap_values(fall_model, X_test=df_for_fall)
    compliance = 0 if compliance_prob < 0.5 else 1

    return {
        'CompleteProb': float(complete_prob),
        'FallProb': float(fall_prob),
        'Compliance': int(compliance),
        'CompleteShapValues': complete_shap_values,
        'FallShapValues': fall_shap_values,
        'ClusterId': int(df.iloc[0]['Cluster']),
        'ClusterValues': cluster_values,
        'Arguments': arguments
    }
    
def add_embeddings(df: pd.DataFrame, case: str) -> pd.DataFrame:
    for i in range(1, ATS_RESOLUTION+1):
        embedding = read_embedding(f'{case}_{i}Ats.csv')
        column = f'{i}Ats'
        df[column] = df[column].replace(to_replace=embedding)
        df[column] = pd.to_numeric(df[column])
    return df

def generate_arguments(df: pd.DataFrame):
    gender = 'kvinde' if df.iloc[0].Gender_Male == 0 else 'mand'
    arguments = list({f'Personen er en {gender}',
                     f'Personen er {int(121 - df.iloc[0].BirthYear)} år gammel',
                     f'Personen har i gennemsnit lånt hjælpemidler i {int(df.iloc[0].LoanPeriod)} dage',
                     f'Personen har {int(df.iloc[0].NumberAts)} hjælpemidler i hjemmet'})
    return arguments

def prepare_data(data: dict) -> pd.DataFrame:
    new_data = {k: [v] for k, v in data.items()}
    new_data_df = pd.DataFrame.from_dict(new_data)
    new_data_df['NumberAts'] = len(new_data_df['Ats'][0].split(","))
    
    df = split_categorical_columns(new_data_df, col='Ats', tag='Ats',
                                   resolution=ATS_RESOLUTION)
    
    df['Gender_Male'] = float(any(df['Gender'] == 1))
    df['Gender_Female'] = float(any(df['Gender'] == 0))
    df = df.drop(['Gender'], axis=1)
    
    cols_ats = [str(i)+'Ats' for i in range(1, ATS_RESOLUTION+1)]
    header_list = ['Gender_Male', 'Gender_Female', 'BirthYear',
                   'Cluster', 'LoanPeriod', 'NumberAts'] + cols_ats
    df = df.reindex(columns=header_list)
    df = df.fillna('0')
    
    citizen_ats = df.filter(regex='((\d+)[Ats])\w+', axis=1).values
    df['Cluster'] = get_cluster(citizen_ats)
    df['Cluster'] = pd.to_numeric(df['Cluster'])
    
    return df

def get_shap_values(model, X_train=None, X_test=None) -> List:
    explainer = shap.TreeExplainer(model, data=X_train)
    shap_values = explainer.shap_values(X_test)
    shap_values_flat = [round(float(val), 3) for val in shap_values[0]]
    shap_values_dict = dict(zip(X_test.columns, shap_values_flat))
    return shap_values_dict

def get_cluster(ats):
    model = read_joblib('kmodes_model.joblib')
    prediction = model.predict(ats)
    return int(prediction[0])

def split_categorical_columns(df: pd.DataFrame, col: str,
                              tag: str, resolution: int) -> pd.DataFrame:
    split = pd.DataFrame(df[col].str.split(pat=",", expand=True))
    split = split.drop(split.iloc[:, resolution:], axis=1)
    split = split.fillna(0)
    df = pd.concat([df, split], axis=1)
    df = df.drop(col, axis=1)
    for i in range(0, resolution):
        try:
            df = df.rename(columns={i: f'{i+1}{tag}'})
        except:
            pass
    return df

def read_embedding(filename: str):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open(f'{dir_path}/embeddings/{filename}', 'r') as f:
        reader = csv.reader(f)
        embedding_dict = {rows[0]:rows[1] for rows in reader}
    return embedding_dict

def read_joblib(filename: str) -> any:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return joblib.load(f'{dir_path}/models/{filename}')

def read_csv(filename:str) -> any:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open(f'{dir_path}/data/{filename}', encoding='utf-8') as f:
        reader = csv.reader(f)
        csv_dict = {rows[0]:rows[1] for rows in reader}
    return csv_dict
    
def read_dataframe(filename:str, converters=None) -> pd.DataFrame:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    df = pd.read_csv(f'{dir_path}/data/{filename}',
                     converters=converters)
    return df

if __name__ == "__main__":
    uvicorn.run(app, port=8000, host="0.0.0.0")